from bioc import DOCUMENT
from biocjson.jsondecoder import parse_doc, parse_passage, parse_sentence, load, loads, BioCJsonDecoderIter
from biocjson.jsonencoder import BioCJSONEncoder, BioCJsonWriter
from biocjson.jsonencoder import dumps, dump

__all__ = ['iterparse', 'iterwrite', 'dumps', 'dump', 'load', 'loads']


def iterparse(file, level=DOCUMENT):
    reader = BioCJsonDecoderIter(file, level)
    return reader


def iterwrite(file, level=DOCUMENT):
    writer = BioCJsonWriter(file, level)
    return writer



import json

import jsonlines

import bioc


def parse_collection(obj: dict) -> bioc.BioCCollection:
    collection = bioc.BioCCollection()
    collection.source = obj['source']
    collection.date = obj['date']
    collection.key = obj['key']
    collection.infons = obj['infons']
    for doc in obj['documents']:
        collection.add_document(parse_doc(doc))
    return collection


def parse_annotation(obj: dict) -> bioc.BioCAnnotation:
    ann = bioc.BioCAnnotation()
    ann.id = obj['id']
    ann.infons = obj['infons']
    ann.text = obj['text']
    for loc in obj['locations']:
        ann.add_location(bioc.BioCLocation(loc['offset'], loc['length']))
    return ann


def parse_relation(obj: dict) -> bioc.BioCRelation:
    rel = bioc.BioCRelation()
    rel.id = obj['id']
    rel.infons = obj['infons']
    for node in obj['nodes']:
        rel.add_node(bioc.BioCNode(node['refid'], node['role']))
    return rel


def parse_sentence(obj: dict) -> bioc.BioCSentence:
    sentence = bioc.BioCSentence()
    sentence.offset = obj['offset']
    sentence.infons = obj['infons']
    sentence.text = obj['text']
    for annotation in obj['annotations']:
        sentence.add_annotation(parse_annotation(annotation))
    for relation in obj['relations']:
        sentence.add_relation(parse_relation(relation))
    return sentence


def parse_passage(obj: dict) -> bioc.BioCPassage:
    passage = bioc.BioCPassage()
    passage.offset = obj['offset']
    passage.infons = obj['infons']
    if 'text' in obj:
        passage.text = obj['text']
    for sentence in obj['sentences']:
        passage.add_sentence(parse_sentence(sentence))
    for annotation in obj['annotations']:
        passage.add_annotation(parse_annotation(annotation))
    for relation in obj['relations']:
        passage.add_relation(parse_relation(relation))
    return passage


def parse_doc(obj: dict) -> bioc.BioCDocument:
    doc = bioc.BioCDocument()
    doc.id = obj['id']
    doc.infons = obj['infons']
    for passage in obj['passages']:
        doc.add_passage(parse_passage(passage))
    for annotation in obj['annotations']:
        doc.add_annotation(parse_annotation(annotation))
    for relation in obj['relations']:
        doc.add_relation(parse_relation(relation))
    return doc


def load(fp, **kwargs) -> bioc.BioCCollection:
    """
    Deserialize fp (a .read()-supporting text file or binary file containing a JSON document) to a BioCCollection object

    Args:
        fp: a file containing a JSON document
        **kwargs:

    Returns:
        BioCCollection: a collection
    """
    obj = json.load(fp, **kwargs)
    return parse_collection(obj)


def loads(s: str, **kwargs) -> bioc.BioCCollection:
    """
    Deserialize s (a str, bytes or bytearray instance containing a JSON document) to a BioCCollection object.

    Args:
        s(str):
        **kwargs:

    Returns:
        BioCCollection: a collection
    """
    obj = json.loads(s, **kwargs)
    return parse_collection(obj)


class BioCJsonDecoderIter(object):
    def __init__(self, file: str, level: int):
        if level not in {bioc.DOCUMENT, bioc.PASSAGE, bioc.SENTENCE}:
            raise ValueError('Unrecognized level: %s' % level)

        self.reader = jsonlines.open(file)
        self.reader_iter = iter(self.reader)
        self.level = level

    def __iter__(self):
        return self

    def __next__(self):
        obj = next(self.reader_iter)
        if self.level == bioc.DOCUMENT:
            return parse_doc(obj)
        elif self.level == bioc.PASSAGE:
            return parse_passage(obj)
        elif self.level == bioc.SENTENCE:
            return parse_sentence(obj)

    def close(self):
        self.reader.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



from typing import List, TextIO, Callable

from bioc.bioc import BioCCollection, BioCDocument, BioCPassage, BioCSentence, BioCAnnotation, BioCRelation, \
    BioCLocation, BioCNode
from bioc.biocitertools import PASSAGE, SENTENCE, DOCUMENT, annotations, relations, sentences
from bioc.decoder import BioCDecoder
from bioc.encoder import BioCEncoder
from bioc.iterdecoder import BioCDecoderIter
from bioc.iterencoder import BioCEncoderIter
from bioc.utils import get_text, pretty_print
from bioc.validator import BioCValidator

__all__ = ['BioCAnnotation', 'BioCCollection', 'BioCDocument', 'BioCLocation', 'BioCNode',
           'BioCPassage', 'BioCRelation', 'BioCSentence', 'load', 'loads', 'dump', 'dumps',
           'iterparse', 'merge', 'validate', 'iterwrite', 'annotations', 'sentences', 'get_text', 'pretty_print']


def dumps(collection: BioCCollection, pretty_print: bool = True) -> str:
    """
    Serialize ``collection`` to a BioC formatted ``str``.

    Args:
        collection: the BioC collection
        pretty_print: enables formatted XML

    Returns:
        a BioC formatted ``str``
    """
    return BioCEncoder(pretty_print=pretty_print).encode(collection)


def dump(collection: BioCCollection, fp: TextIO, pretty_print: bool = True):
    """
    Serialize ``collection`` as a BioC formatted stream to ``fp``.

    Args:
        collection: the BioC collection
        fp: a ``.write()``-supporting file-like object
        pretty_print: enables formatted XML
    """
    fp.write(BioCEncoder(pretty_print=pretty_print).encode(collection))


def load(fp: TextIO) -> BioCCollection:
    """
    Deserialize ``fp`` to a BioC collection object.

    Args:
        fp: a ``.read()``-supporting file-like object containing a BioC collection

    Returns:
         a object of BioCollection
    """
    return BioCDecoder().decode(fp)


def loads(s: str) -> BioCCollection:
    """
    Deserialize ``s`` to a BioC collection object.

    Args:
        s: a ``str`` instance containing a BioC collection

    Returns:
        an object of BioCollection
    """
    return BioCDecoder().decodes(s)


def validate(collection, onerror: Callable[[str, List], None] = None):
    BioCValidator(onerror).validate(collection)


def merge(output: str, *input: str):
    """
    Merge multiple BioC files into one.

    Args:
        output: output BioC file name
        input: input BioC file names
    """
    collection = None
    for src in input:
        with open(src) as fp:
            tmp = load(fp)
        if collection is None:
            collection = tmp
        else:
            for document in tmp.documents:
                collection.add_document(document)
    with open(output, 'w') as fp:
        dump(collection, fp)


def iterparse(file) -> BioCDecoderIter:
    if not isinstance(file, str):
        file = str(file)
    parser = BioCDecoderIter(file)
    return parser


def iterwrite(file, collection=None) -> BioCEncoderIter:
    if not isinstance(file, str):
        file = str(file)
    writer = BioCEncoderIter(file, collection)
    return writer



import json
import jsonlines
import bioc


def dumps(obj, **kwargs) -> str:
    """
    Serialize a BioC ``obj`` to a JSON formatted ``str``.
    """
    return json.dumps(obj, cls=BioCJSONEncoder, **kwargs)


def dump(obj, fp, **kwargs):
    """
    Serialize obj as a JSON formatted stream to fp (a .write()-supporting file-like object)
    """
    return json.dump(obj, fp, cls=BioCJSONEncoder, **kwargs)


class BioCJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, bioc.BioCLocation):
            return {
                'offset': obj.offset,
                'length': obj.length,
            }
        elif isinstance(obj, bioc.BioCAnnotation):
            return {
                'id': obj.id,
                'infons': obj.infons,
                'text': obj.text,
                'locations': [self.default(l) for l in obj.locations],
            }
        elif isinstance(obj, bioc.BioCNode):
            return {
                'refid': obj.refid,
                'role': obj.role,
            }
        elif isinstance(obj, bioc.BioCRelation):
            return {
                'id': obj.id,
                'infons': obj.infons,
                'nodes': [self.default(n) for n in obj.nodes]
            }
        elif isinstance(obj, bioc.BioCSentence):
            return {
                'offset': obj.offset,
                'infons': obj.infons,
                'text': obj.text,
                'annotations': [self.default(a) for a in obj.annotations],
                'relations': [self.default(r) for r in obj.relations],
            }
        elif isinstance(obj, bioc.BioCPassage):
            return {
                'offset': obj.offset,
                'infons': obj.infons,
                'text': obj.text,
                'sentences': [self.default(s) for s in obj.sentences],
                'annotations': [self.default(a) for a in obj.annotations],
                'relations': [self.default(r) for r in obj.relations],
            }
        elif isinstance(obj, bioc.BioCDocument):
            return {
                'id': obj.id,
                'infons': obj.infons,
                'passages': [self.default(p) for p in obj.passages],
                'annotations': [self.default(a) for a in obj.annotations],
                'relations': [self.default(r) for r in obj.relations],
            }
        elif isinstance(obj, bioc.BioCCollection):
            return {
                'source': obj.source,
                'date': obj.date,
                'key': obj.key,
                'infons': obj.infons,
                'documents': [self.default(d) for d in obj.documents],
            }
        return json.JSONEncoder.default(self, obj)


class BioCJsonWriter(object):
    def __init__(self, file, level):
        if level not in {bioc.DOCUMENT, bioc.PASSAGE, bioc.SENTENCE}:
            raise ValueError('Unrecognized level: %s' % level)

        self.writer = jsonlines.open(file, 'w')
        self.level = level

    def write(self, obj):
        if self.level == bioc.DOCUMENT and not isinstance(obj, bioc.BioCDocument):
            raise ValueError
        if self.level == bioc.PASSAGE and not isinstance(obj, bioc.BioCPassage):
            raise ValueError
        if self.level == bioc.SENTENCE and not isinstance(obj, bioc.BioCSentence):
            raise ValueError
        self.writer.write(BioCJSONEncoder().default(obj))

    def close(self):
        self.writer.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()